
/*  
 * $Id: cache.c,v 1.15 2005/02/03 17:52:31 isv Exp $
 *
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2004
 * See License.txt for details
 *
 */

#include "module.h"

static KLIST_HEAD(files_cache);
static int max_cache_files = -1;
static lock_declare(cache_lock);


void Monitor_cache_init(int max_files)
{
	if (max_cache_files == -1){
		KINIT_LIST_HEAD(&files_cache);
		lock_init(&cache_lock,"kavmonitor cache");
	} else {
		Monitor_cache_finit();
	}

	max_cache_files = max_files;
}

void Monitor_cache_finit(void)
{
	struct list_head        * tmp;
	cache_entry             * hash;

again1:
	write_lock (&cache_lock);

        list_for_each(tmp, &files_cache) {
                hash = list_entry(tmp,cache_entry,list);
                list_del(&hash->list);
		
		write_unlock (&cache_lock);

                MEM_FREE(hash);
                goto again1;
        }
	
	
	write_unlock (&cache_lock);	
}

void Monitor_cache_add(const file_info * file)
{
	struct list_head        * tmp;
	cache_entry		* hash, * min_hash;
	int			  files_count;

	if (!max_cache_files) {
#if defined(__linux__)
		printk("Monitor_cache_add: function has been called when max_cache_files=0\n");
#elif defined (__FreeBSD__)
		uprintf("Monitor_cache_add: function has been called when max_cache_files=0\n");
#endif		
		return;
	}
	
	files_count	= 0;
	min_hash	= NULL;
	
	write_lock (&cache_lock);

	list_for_each(tmp, &files_cache) {
		hash = list_entry(tmp,cache_entry,list);
		if (FILE_EQUAL(&hash->file, file)){
				
			hash->count++;
			write_unlock (&cache_lock);
			return;
                }

		files_count ++;
                if (!min_hash) min_hash = hash; else {
			if (min_hash->count > hash->count) min_hash = hash;
		}

                if (!min_hash) min_hash = hash; else {
			if (min_hash->count > hash->count) min_hash = hash;
		}

	}

	if (min_hash && files_count >= max_cache_files) {
		hash = min_hash;

		list_del(&hash->list);

		memcpy(&hash->file,file,sizeof(file_info));
		hash->count++;
		write_unlock (&cache_lock);

		list_add(&hash->list,&files_cache);

		return;

	}
	
	write_unlock (&cache_lock);

	hash = (cache_entry *) MEM_ALLOC(sizeof(cache_entry));

	write_lock (&cache_lock);
	memcpy(&hash->file,file,sizeof(file_info));

	hash->count = 0;
	list_add(&hash->list,&files_cache);

	write_unlock (&cache_lock);
}

void Monitor_cache_del(const file_info * file)
{
	struct list_head        * tmp;
	cache_entry		* hash = NULL;
	int			  need_free;

again:	
	need_free = 0;

	write_lock (&cache_lock);

	list_for_each(tmp, &files_cache) {
		hash = list_entry(tmp,cache_entry,list);
		if (FILE_EQUAL(&hash->file, file)){
			list_del(&hash->list);
			need_free = 1;
			break;
                }
	}
	write_unlock (&cache_lock);
	
	if (need_free) {
		MEM_FREE(hash);
		goto again;
	}
}


int Monitor_cached(const file_info * file)
{

	struct list_head        * tmp;
	cache_entry		* hash;

	read_lock (&cache_lock);

	list_for_each(tmp, &files_cache) {
                hash = list_entry(tmp,cache_entry,list);
		if (FILE_EQUAL(&hash->file, file)) {
			hash->count++;
			read_unlock (&cache_lock);
			return 1;
                }
        }

	read_unlock (&cache_lock);
	return 0;
}


