
/*  
 * $Id: kernel.c,v 1.53 2005/02/16 14:26:05 isv Exp $
 *
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2004
 * See License.txt for details
 *
 */

#include "module.h"

static void monitor_finit_trusted(void);
static void file_checked(TASK_INFO u_short queue_id,unsigned long state ,char * viruses);
static void monitor_shutdown(TASK_INFO_);
static void file_answer_sign(TASK_INFO u_short queue_id, u_short result);
static void message_answer(TASK_INFO u_short queue_id,enum monitor_msg_action action);
static int  kernel_send_msg(TASK_INFO uid_t uid, pid_t pid, enum monitor_msg_type msg_type,
				const char * msg,enum queue_entry_types queue_type,
				enum monitor_msg_action * action, int * err);

static void add_exec_exclude(const char * prog,struct list_head * hash_list,
		      struct list_head * str_list);

static int kernel_exec_denied(const char * filename,struct list_head * hash_list,
		      struct list_head * str_list);

monitor_callbacks m_callbacks={
		file_check_request:NULL,
		file_check_answer:file_checked,
		monitor_shutdown:monitor_shutdown,
		file_check_sign: NULL,
		file_answer_sign:file_answer_sign,
		monitor_msg:NULL,
		monitor_answer_msg:message_answer
	};


static lock_declare(trusted_lock);
static KLIST_HEAD(trusted_progs);
static KLIST_HEAD(trusted_groups);

static exec_network_deny	exec_denies;

monitor_config m_config;


void Monitor_kernel_init(void)
{
	lock_init(&trusted_lock,"kavmonitor kernel");

	KINIT_LIST_HEAD(&exec_denies.net_hash_excludes);
	KINIT_LIST_HEAD(&exec_denies.net_str_excludes);

	KINIT_LIST_HEAD(&exec_denies.stack_hash_excludes);
	KINIT_LIST_HEAD(&exec_denies.stack_str_excludes);
	KINIT_LIST_HEAD(&trusted_progs);

	m_config.exec_network_prot = 0;
	m_config.exec_stack_prot = 0;
	m_config.cache_enable = 0;

	Monitor_callbacks_init(&m_callbacks);
}

void Monitor_kernel_finit(void)
{
	struct list_head        * tmp;
	hash_entry		* hash;
	str_entry 		* str;

again1:
	write_lock(&trusted_lock);

	list_for_each(tmp, &exec_denies.net_hash_excludes) {
		hash = list_entry(tmp,hash_entry,list);
		list_del(&hash->list);
		write_unlock(&trusted_lock);

		MEM_FREE(hash);
		goto again1;
	}

	write_unlock(&trusted_lock);


again2:	
	write_lock(&trusted_lock);

	list_for_each(tmp, &exec_denies.net_str_excludes) {
		str = list_entry(tmp,str_entry,list);
		list_del(&str->list);
		write_unlock(&trusted_lock);

		MEM_FREE(str->path);
		MEM_FREE(str);
		goto again2;
	}

	write_unlock(&trusted_lock);
	
again3:	
	write_lock(&trusted_lock);

	list_for_each(tmp, &exec_denies.stack_hash_excludes) {
		hash = list_entry(tmp,hash_entry,list);
		list_del(&hash->list);
		write_unlock(&trusted_lock);

		MEM_FREE(hash);
		goto again3;
	}

	write_unlock(&trusted_lock);

again4:
	write_lock(&trusted_lock);

	list_for_each(tmp, &exec_denies.stack_str_excludes) {
		str = list_entry(tmp,str_entry,list);
		list_del(&str->list);
		write_unlock(&trusted_lock);

		MEM_FREE(str->path);
		MEM_FREE(str);
		goto again4;
	
	}

	write_unlock(&trusted_lock);


	monitor_finit_trusted();
	Monitor_cache_finit();
}

void monitor_finit_trusted(void)
{
	task_entry		* task;
	group_entry		* group;
	struct list_head        * tmp;


again1:
	write_lock(&trusted_lock);

	list_for_each(tmp, &trusted_progs) {
		task = list_entry(tmp,task_entry,list);
		list_del(&task->list);

		write_unlock(&trusted_lock);
		MEM_FREE(task);
		goto again1;
	}


	write_unlock(&trusted_lock);

again2:
	write_lock(&trusted_lock);


	list_for_each(tmp, &trusted_groups) {
		group = list_entry(tmp,group_entry,list);
		list_del(&group->list);

		write_unlock(&trusted_lock);

		MEM_FREE(group);
		goto again2;
	}

	write_unlock(&trusted_lock);
}	




void Monitor_kernel_register(pid_t pid, const file_info *  file)
{
	task_entry		 * task;
	struct list_head	 * tmp;

	write_lock(&trusted_lock);

	list_for_each(tmp, &trusted_progs) {
		task = list_entry(tmp,task_entry,list);
		if (task->pid == pid) {
			write_unlock(&trusted_lock);
			return;
		}
	}

	write_unlock(&trusted_lock);

	task = (task_entry *) MEM_ALLOC(sizeof(task_entry));
	task->pid = pid;

	memcpy(&task->file, file, sizeof(file_info));

	write_lock(&trusted_lock);
	list_add(&task->list,&trusted_progs);
	write_unlock(&trusted_lock);
}


void Monitor_kernel_register_group(pid_t pgrp)
{
	group_entry	 * group;
	struct list_head * tmp;

	write_lock(&trusted_lock);

	list_for_each(tmp, &trusted_groups) {
		group = list_entry(tmp,group_entry,list);
		if (group->pgrp == pgrp) {
			write_unlock(&trusted_lock);
			return;
		}
	}

	write_unlock(&trusted_lock);

	group = (group_entry *) MEM_ALLOC(sizeof(group_entry));
	group->pgrp = pgrp;

	write_lock(&trusted_lock);
	list_add(&group->list,&trusted_groups);
	write_unlock(&trusted_lock);
}



void Monitor_kernel_unregister_group(pid_t pgrp)
{
	group_entry	 * group = NULL;
	struct list_head * tmp;
	int		   need_free = 0;

	write_lock(&trusted_lock);

	list_for_each(tmp, &trusted_groups) {
		group = list_entry(tmp,group_entry,list);
		if (group->pgrp == pgrp) {
			list_del(&group->list);
			need_free = 1;
			break;
		}
	}

	write_unlock(&trusted_lock);

	if (need_free) {
		MEM_FREE(group);
	}
}


void Monitor_kernel_unregister(pid_t pid)
{
	struct list_head        * tmp;
	task_entry		* task = NULL;
	int			  need_free = 0;
	
	write_lock(&trusted_lock);
	list_for_each(tmp, &trusted_progs) {
		task = list_entry(tmp,task_entry,list);
		if (task->pid == pid){
			list_del(&task->list);
			need_free = 1;
			break;
		}
	}

	write_unlock (&trusted_lock);

	if (need_free) {
		MEM_FREE(task);
	}
}

int Monitor_kernel_trusted(pid_t pid, pid_t pgrp)
{	
	struct list_head        * tmp;
	task_entry		* task;
	group_entry		* group;

	read_lock(&trusted_lock);

	list_for_each(tmp, &trusted_groups) {
		group = list_entry(tmp,group_entry,list);
			if (group->pgrp == pgrp) {
				read_unlock(&trusted_lock);
				return 1;
		}
	}

	list_for_each(tmp, &trusted_progs) {
		task = list_entry(tmp,task_entry,list);
			if (task->pid == pid) {
				read_unlock(&trusted_lock);
				return 1;
		}
	}

	read_unlock(&trusted_lock);
	return 0;
}


int  Monitor_kernel_trusted_1(const file_info * file)
{
	struct list_head        * tmp;
	task_entry		* task;

	read_lock(&trusted_lock);

	list_for_each(tmp, &trusted_progs) {
		task = list_entry(tmp,task_entry,list);
			if (FILE_EQUAL(&task->file, file)){
				read_unlock(&trusted_lock);
				return 1;
		}
	}

	read_unlock(&trusted_lock);
	return 0;

}


void Monitor_kernel_parse(TASK_INFO char * data)
{
	Monitor_Request_Ready(TASK NULL, NULL,data);
}

int Monitor_kernel_check_file(uid_t uid, pid_t pid,const char * filename,
				const file_info * info,
				enum monitor_file_type filetype,int flags, int * err)
{
	u_short	  queue_id;
	long	  answer = (long )MONITOR_ACCESS_ACCEPT;
	char	* request_data;
	int	  request_len;

	*err = 0;
	
	if (m_config.cache_enable) {
		
		if ((KAVMON_IS_WRITTEN(flags) && filetype ==MONITOR_CLOSE_FILE) ||
			filetype == MONITOR_UNLINK_FILE ||
			filetype == MONITOR_RENAME_FILE) {

			Monitor_cache_del(info);
		} else if (Monitor_cached(info)) return 0;
	}
 
	queue_id = Monitor_queue_get_id();

	request_data = Monitor_make_file_request(uid,pid,
					queue_id,filename,filetype,flags,
					&request_len);
	DEBUG_MSG("%s: pid=%d queue_id=%d Scanning of filename=%s has been started",__FUNCTION__,pid,queue_id,filename);


	*err = Monitor_queue_add (QUEUE_MONITOR, uid,pid,request_data, request_len, &answer, queue_id);
	DEBUG_MSG("%s: pid=%d queue_id=%d Scanning of filename=%s has been done answer=%d",__FUNCTION__,
			pid,queue_id,filename,answer);

	if (*err) return *err;

	if (answer == MONITOR_ACCESS_ACCEPT && m_config.cache_enable && 
			filetype != MONITOR_CLOSE_FILE  && filetype != MONITOR_UNLINK_FILE && 
			filetype != MONITOR_RENAME_FILE) {
		 Monitor_cache_add(info);
	}
#if defined(__linux__)
	return (answer == MONITOR_ACCESS_ACCEPT)?0:-EACCES;
#elif defined(__FreeBSD__)
	return (answer == MONITOR_ACCESS_ACCEPT)?0:EACCES;
#endif
}

int Monitor_kernel_check_file_sign(uid_t uid, pid_t pid,const char * filename, int * err)
{
	char	* request_data;
	long	  answer;
	u_short	  queue_id;
	int	  request_len;
	
	*err = 0;
	
	queue_id = Monitor_queue_get_id();
	request_data = Monitor_make_sign_request(queue_id,filename,&request_len);
	answer = -1; 
	
	*err = Monitor_queue_add (QUEUE_MONITOR,uid,pid,request_data, request_len, &answer, queue_id);

	return answer;
}

int Monitor_kernel_exploit_alert(TASK_INFO uid_t uid, pid_t pid, enum monitor_msg_type msg_type,
				const char * msg,enum monitor_msg_action * action, int * err)
{

	return kernel_send_msg(TASK uid,pid,msg_type,msg,QUEUE_EXPLOIT, action, err);

}

void Monitor_kernel_add_exec_net_exclude(const char * prog)
{
	add_exec_exclude(prog,&exec_denies.net_hash_excludes,
			 &exec_denies.net_str_excludes);
	
}

void Monitor_kernel_add_exec_stack_exclude(const char * prog)
{
	add_exec_exclude(prog,&exec_denies.stack_hash_excludes,
			 &exec_denies.stack_str_excludes);
}

int Monitor_kernel_exec_net_denied(const char * filename)
{
	return kernel_exec_denied(filename,&exec_denies.net_hash_excludes,
			 &exec_denies.net_str_excludes);
	
}

int Monitor_kernel_exec_stack_denied(const char * filename)
{
	return kernel_exec_denied(filename,&exec_denies.stack_hash_excludes,
			 &exec_denies.stack_str_excludes);
}

void file_checked(TASK_INFO u_short queue_id,unsigned long file_status ,char * viruses)
{

	Monitor_queue_set_answer(queue_id, (long )file_status);
	Monitor_queue_wakeup(queue_id);
}

void monitor_shutdown(TASK_INFO_)
{
	DEBUG_MSG("%s:",__FUNCTION__);

	Monitor_module_stop(TASK_);
}

void file_answer_sign(TASK_INFO u_short queue_id,u_short result)
{
	Monitor_queue_set_answer(queue_id, (long )result);
	Monitor_queue_wakeup(queue_id);
}

void message_answer(TASK_INFO u_short queue_id,enum monitor_msg_action action)
{
	Monitor_queue_set_answer(queue_id, (long )action);
	Monitor_queue_wakeup(queue_id);
}


int kernel_send_msg(TASK_INFO uid_t uid, pid_t pid, enum monitor_msg_type msg_type,
				const char * msg,enum queue_entry_types queue_type,
				enum monitor_msg_action *action, int * err)
{
	char	* request_data;
	int	  request_len;
	long	  answer;
	u_short	  queue_id;
	u_short	  result;

	*err   = 0;
	result = 0;

	queue_id = Monitor_queue_get_id();
	request_data = Monitor_make_msg_request(uid,pid,queue_id,msg_type,msg,&request_len);

	answer = -1;
	
	*err = Monitor_queue_add (queue_type, uid,pid,request_data, request_len, &answer, queue_id);
	
	*action = (enum monitor_msg_action) result;
	
	return 0;
}


void add_exec_exclude(const char * prog,struct list_head * hash_list,
		      struct list_head * str_list)
{
	int i, wildcards;
	
	for(i=0, wildcards=0;prog[i]!=0;i++) 
		if (prog[i]=='*' || prog[i]=='[' || prog[i]==']' || 
		    prog[i]=='.' || prog[i]=='^')  {
			wildcards =1;
			break;
		}

	if (wildcards) {
		str_entry * str;

		str = (str_entry *) MEM_ALLOC(sizeof(str_entry)+1);
		str->path = (char *) MEM_ALLOC(strlen(prog)+1);
		strcpy(str->path,prog);


		write_lock(&trusted_lock);
		list_add(&str->list,str_list);
		
	} else {
		hash_entry * hash;
		struct MD5Context ctx;

		hash = (hash_entry *) MEM_ALLOC(sizeof(hash_entry)+1);

		MD5Init(&ctx);
		MD5Update(&ctx, (unsigned char const *)prog,strlen(prog));
		MD5Final(hash->hash, &ctx);
		
		write_lock(&trusted_lock);
		list_add(&hash->list,hash_list);
	}

	write_unlock(&trusted_lock);
}

int kernel_exec_denied(const char * filename,struct list_head * hash_list,
		      struct list_head * str_list)
{
	int			  i;
	struct MD5Context	  ctx;
	unsigned char		  name_hash[16];
	struct list_head        * tmp_list;
	hash_entry		* hash;
	str_entry 		* str;
	const char		* pname;

	tmp_list = hash_list;

	if (tmp_list !=tmp_list->next) {
		MD5Init(&ctx);
		MD5Update(&ctx, (unsigned char const *)filename,strlen(filename));
		MD5Final(name_hash, &ctx);


		list_for_each(tmp_list, hash_list) {
			hash = list_entry(tmp_list,hash_entry,list);
			if (!memcmp(hash->hash,name_hash,sizeof(name_hash))) {
				return 0;
			}
		}
	}

	tmp_list = str_list;

	if (tmp_list!= tmp_list->next) {
		for(i = strlen(filename); i>0; i--) if (filename[i]=='/') break;
		pname = filename+i+1;
	
		list_for_each(tmp_list, str_list) {
			str = list_entry(tmp_list,str_entry,list);
			if (str->path[0]=='/') {
				if (__fnmatch(filename,str->path)) return 0;
			} else {
				if (__fnmatch(pname,str->path)) return 0;
			}

		}
	}

	return 1;
}

