
/*
 * $Id: interceptor.c,v 1.64 2005/05/20 12:15:44 isv Exp $
 *
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2004
 * See License.txt for details
 *
 */


#include "module.h"
#include <linux/sched.h>

typedef struct {
	struct file		* file;
	struct task_struct	* tsk;
	struct list_head	  list;

} file_t;

#ifdef NO_DO_EXECVE
int (*do_execve_)(char *, char **, char **, struct pt_regs *) = 
		(int (*)(char *, char **, char **, struct pt_regs *)) DO_EXECVE_ADDR;
#else
int (*do_execve_)(char *, char **, char **, struct pt_regs *) = do_execve;
#endif


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,48))
#define DENTRY_PATH(file) get_path_by_dentry (file->f_dentry, NULL)
#else
#define DENTRY_PATH(file) get_path_by_dentry (file->f_dentry, file->f_vfsmnt)
#endif

static int file_really_written(struct file * file);
static int monitor_exec_net_denied(const char * filename, int *err);
static int monitor_exec_stack_denied(const char * filename, int * err);

/* old prototypes */
static asmlinkage int (*old_sys_close) (unsigned int fd);
static asmlinkage int (*old_sys_execve)(struct pt_regs regs);
static asmlinkage int (*old_sys_open)  (const char *, int, int);
static asmlinkage int (*old_sys_unlink)(const char *);
static asmlinkage int (*old_sys_rename)(const char *, const char *);
static asmlinkage ssize_t (*old_sys_write)(unsigned int , const char * , size_t );
static asmlinkage int (*old_sys_fork)(struct pt_regs);
static asmlinkage long (*old_sys_exit)(int );
#ifdef LINUX_KERNEL26
static asmlinkage long (*old_sys_exit_group)(int );
#endif
static asmlinkage long (*old_sys_setpgid)(pid_t, pid_t);


static asmlinkage int new_sys_open (const char * u_file, int flags, int mode);
static asmlinkage int new_sys_execve(struct pt_regs regs);
static asmlinkage int new_sys_close (unsigned int fd);
static asmlinkage int new_sys_unlink(const char * u_file);
static asmlinkage int new_sys_rename(const char *u_old_file, const char * u_new_file);
static asmlinkage ssize_t new_sys_write(unsigned int fd, const char * buf, size_t count);
static asmlinkage int new_sys_fork(struct pt_regs regs);
static asmlinkage long new_sys_exit(int error_code);
#ifdef LINUX_KERNEL26
static asmlinkage long new_sys_exit_group(int error_code);
#endif
static asmlinkage long new_sys_setpgid(pid_t pid, pid_t pgid);
static void __exit_tsk(void);


static struct list_head written_files;
static lock_declare(written_files_lock);

extern void ** sys_calls_table;

int Monitor_intercept_calls (void)
{
	/* open() */
	old_sys_open = sys_calls_table[__NR_open];
	sys_calls_table[__NR_open] = new_sys_open;

	/* execve() */
	old_sys_execve = sys_calls_table[__NR_execve];
	sys_calls_table[__NR_execve] = new_sys_execve;

	/* write */
	old_sys_write = sys_calls_table[__NR_write];
	sys_calls_table[__NR_write] = new_sys_write;
	
	/* close() */
	old_sys_close = sys_calls_table[__NR_close];
	sys_calls_table[__NR_close] = new_sys_close;

	/* unlink */
	old_sys_unlink = sys_calls_table[__NR_unlink];
	sys_calls_table[__NR_unlink] = new_sys_unlink;
	
	/* rename */
	old_sys_rename = sys_calls_table[__NR_rename];
	sys_calls_table[__NR_rename] = new_sys_rename;

	/* fork */
	old_sys_fork = sys_calls_table[__NR_fork];
	sys_calls_table[__NR_fork] = new_sys_fork;
	
	/* exit */
	old_sys_exit = sys_calls_table[__NR_exit];
	sys_calls_table[__NR_exit] = new_sys_exit;

#ifdef LINUX_KERNEL26
	/* exit */
	old_sys_exit_group = sys_calls_table[__NR_exit_group];
	sys_calls_table[__NR_exit_group] = new_sys_exit_group;

#endif	

	/* setpgid */
	old_sys_setpgid = sys_calls_table[__NR_setpgid];
	sys_calls_table[__NR_setpgid] = new_sys_setpgid;
	
	
	INIT_LIST_HEAD(&written_files);
	lock_init(&written_files_lock,"kavmonitor write lock");
	
    return 0;
}


int Monitor_restore_calls (void)
{
	struct list_head        * f_list;
	file_t			* file;


	/* Restore open() call */
	if ( sys_calls_table[__NR_open] != new_sys_open
	|| sys_calls_table[__NR_execve] != new_sys_execve
	|| sys_calls_table[__NR_close] != new_sys_close 
	|| sys_calls_table[__NR_rename] != new_sys_rename 
	|| sys_calls_table[__NR_unlink] != new_sys_unlink 
	|| sys_calls_table[__NR_fork] != new_sys_fork
	|| sys_calls_table[__NR_exit] != new_sys_exit
#ifdef LINUX_KERNEL26
	|| sys_calls_table[__NR_exit_group] != new_sys_exit_group
#endif	
	
	|| sys_calls_table[__NR_write] != new_sys_write 
	||  sys_calls_table[__NR_setpgid] != new_sys_setpgid) return 1;


	sys_calls_table[__NR_open] = old_sys_open;
	sys_calls_table[__NR_execve] = old_sys_execve;
	sys_calls_table[__NR_close] = old_sys_close;
	sys_calls_table[__NR_unlink] = old_sys_unlink;
	sys_calls_table[__NR_rename] = old_sys_rename;
	sys_calls_table[__NR_fork] = old_sys_fork;
	sys_calls_table[__NR_exit] = old_sys_exit;
#ifdef LINUX_KERNEL26
	sys_calls_table[__NR_exit_group] = old_sys_exit_group;
#endif	
	sys_calls_table[__NR_write] = old_sys_write;
	sys_calls_table[__NR_setpgid] = old_sys_setpgid;


	/* clean all written files are in the queue */

again:	
	write_lock (&written_files_lock);

	list_for_each(f_list, &written_files) {
		file = list_entry(f_list,file_t,list);
		list_del(&file->list);
		write_unlock (&written_files_lock);
		fput(file->file);
		MEM_FREE(file);
		goto again;
	}
		
	write_unlock (&written_files_lock);

	
	return 0;
}



int Process_trusted(struct task_struct * tsk)
{
	struct task_struct * tmp_tsk = NULL;

	if (Monitor_kernel_trusted(tsk->pid,get_pgrp(tsk))) {
		
		return 1;
	}

	read_lock(&tasklist_lock);

#ifdef LINUX_KERNEL26
	for(tmp_tsk=tsk->parent; tmp_tsk->pid > 1; tmp_tsk=tmp_tsk->parent){

#else

#ifndef FIX_TASK_STRUCT			
	for(tmp_tsk=tsk->p_pptr; tmp_tsk->pid > 1; tmp_tsk=tmp_tsk->p_pptr){
#else
	for(tmp_tsk=tsk->parent; tmp_tsk->pid > 1; tmp_tsk=tmp_tsk->parent){
#endif		
			
#endif			
			if (Monitor_kernel_trusted(tmp_tsk->pid, get_pgrp(tmp_tsk))) {

			read_unlock(&tasklist_lock);
			return 1;
		}

	}


	read_unlock(&tasklist_lock);

	return 0;
}

int file_really_written(struct file * file)
{
	struct list_head        * f_list;
	file_t			* write_file;
	
	write_lock (&written_files_lock);

	list_for_each(f_list, &written_files) {
		write_file = list_entry(f_list,file_t,list);
		
		if (write_file->file == file && write_file->tsk == current) {
			list_del(&write_file->list);
			write_unlock (&written_files_lock);

			fput(write_file->file);
			MEM_FREE(write_file);
			return 1;
		}
	}
		
	write_unlock (&written_files_lock);


	return 0;
}



int monitor_exec_net_denied(const char * filename, int *err)
{
	struct		  	files_struct * files;
	struct file	*	std_in;
	enum monitor_msg_action	action;
	int			denied;
	char		*       exec_path;


	*err		= 0;
	denied		= 0;
	exec_path	= NULL;
	
	files = current->files;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,48) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0) )
	write_lock(&files->file_lock);
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	spin_lock(&files->file_lock);
#endif		
	if (files->max_fds>=1) {
		std_in = files->fd[0];	
		if (std_in && std_in->f_dentry->d_inode && std_in->f_dentry->d_inode->i_sock) {
			file_info	  f_info;
			
			exec_path = get_exec_path(current,&f_info);

			if (!exec_path) goto Done;

			if (Monitor_kernel_exec_net_denied(exec_path)){
				DEBUG_MSG("%s tried to run a shell with stdin %s"
						" on network socket",__FUNCTION__,
						filename);
				Monitor_kernel_exploit_alert(current, current->uid, 
							current->pid,
							MONITOR_ALERT_EXEC_NETWORK,
							filename,&action, err);

				denied = 1;

				if (err) goto Done;
				
				switch(action) {
					case MONITOR_DENY_ACT:
						break;
					case MONITOR_TERM_ACT:
						force_sig(SIGTERM, current);
						break;	
					case MONITOR_KILL_ACT:
						force_sig(SIGKILL, current);
						break;	
					case MONITOR_ABORT_ACT:
						force_sig(SIGSEGV, current);
						break;	
					case MONITOR_INFO_ACT:
						denied = 0;
				}
				goto Done;
			}
		}
	}

Done:		
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,48) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0) )
	write_unlock(&files->file_lock);
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	spin_unlock(&files->file_lock);
#endif		

	if (exec_path) MEM_FREE(exec_path);
	
	return denied;
}


int monitor_exec_stack_denied(const char * filename,int *err)
{
	char * exec_path;
	enum monitor_msg_action	action;
	int			denied;
	file_info		f_info;

	exec_path	= get_exec_path(current,&f_info);
	*err		= 0;
	
	if (!exec_path) return 0;

	if (Monitor_kernel_exec_stack_denied(exec_path)){
		Monitor_kernel_exploit_alert(current, current->uid, 
					     current->pid,
					     MONITOR_ALERT_EXEC_STACK,
					     filename,&action, err);
		denied = 1;

		if (err) {
			MEM_FREE(exec_path);
			return denied;
		}
		
		switch(action) {
			case MONITOR_DENY_ACT:
				break;
			case MONITOR_TERM_ACT:
				force_sig(SIGTERM, current);
				break;	
			case MONITOR_KILL_ACT:
				force_sig(SIGKILL, current);
				break;	
			case MONITOR_ABORT_ACT:
				force_sig(SIGSEGV, current);
				break;	
			case MONITOR_INFO_ACT:
				denied = 0;
		}
					
		MEM_FREE(exec_path);
		return denied;
	}
	MEM_FREE(exec_path);
	return 0;
}


asmlinkage int new_sys_open (const char * u_file, int flags, int mode)
{
	int		   err;
	char		 * k_filename;
	file_info	   f_info;

	if (!monitor_started || Process_trusted(current)) goto Done;
	
	get_file_from_userspace (u_file,  &f_info, &k_filename);
	
	if ( KAVMON_IS_WRITTEN (flags) && k_filename && Monitor_kernel_trusted_1(&f_info)){

		MEM_FREE(k_filename);
		return -EPERM;

	}

	if ( k_filename)  {

		err = Monitor_kernel_check_file(current->uid,current->pid,k_filename,
					&f_info,
					MONITOR_OPEN_FILE,
					flags, &err);

		if ( err ) {
			    MEM_FREE (k_filename);				
			    return err;
			}
	}

	
	if (k_filename) MEM_FREE (k_filename);	
Done:
	err = old_sys_open (u_file, flags, mode);
	
	return err;
}


asmlinkage int new_sys_execve(struct pt_regs regs)
{
	int		  err;
	char		* u_filename = (char*) regs.ebx;
	char		* k_filename = NULL;
	file_info	  f_info;


	if (!monitor_started || Process_trusted(current)) goto Done;

	err = 0;

	get_file_from_userspace (u_filename, &f_info, &k_filename);

	
	if (k_filename){

		if (m_config.exec_stack_prot && exploit_started) {
			int  eip;

			eip = KSTK_EIP(current);

			/* checking if the exec syscall was made from stack code */
			if ((eip & 0xFF800000) == 0xBF800000 ||
					(eip >= PAGE_OFFSET - _STK_LIM && eip < PAGE_OFFSET)) {

				if (monitor_exec_stack_denied(k_filename,&err)) {
					if (!err) err = -EACCES;
					goto Done1;
				}
			
			}
		}
		
		if (m_config.exec_network_prot && exploit_started) {

			if (monitor_exec_net_denied(k_filename,&err)) {
				if (!err) err = -EACCES;
				goto Done1;
			}
		}
		
		if (monitor_started) {
			
		err = Monitor_kernel_check_file(current->uid,current->pid,
					k_filename,
					&f_info,
					MONITOR_EXEC_FILE,0, &err);
					}

Done1:
		if (k_filename) MEM_FREE (k_filename);

		if (err) return err;
	}

Done:

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	{
		k_filename = getname((char *) regs.ebx);
		err = PTR_ERR(k_filename);
		if (IS_ERR(k_filename))
			goto out;
		err = do_execve_(k_filename,
                        (char __user * __user *) regs.ecx,
                        (char __user * __user *) regs.edx,
                        &regs);
		
		if (err == 0) {
			current->ptrace &= ~PT_DTRACE;
	                /* Make sure we don't return using sysenter.. */
        	        set_thread_flag(TIF_IRET);
		}
		
		putname(k_filename);
out:
		return err;
	}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		k_filename = getname((char *) regs.ebx);
		err = PTR_ERR(k_filename);
		if (IS_ERR(k_filename))
			goto out;
		err = do_execve_(k_filename, (char **) regs.ecx, (char **) regs.edx, &regs);
		if (err == 0)
			current->ptrace &= ~PT_DTRACE;
		putname(k_filename);
out:
		return err;
	}
	#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,20)
	{
		#ifdef __SMP__
			lock_kernel();
		#endif
		k_filename = getname((char *) regs.ebx);
		err = PTR_ERR(k_filename);
		if (IS_ERR(k_filename))
			goto out;
		err = do_execve_(k_filename, (char **) regs.ecx, (char **) regs.edx, &regs);
		if (err == 0)
			current->ptrace &= ~PT_DTRACE;
		putname(k_filename);
out:
		#ifdef __SMP__
			unlock_kernel();
		#endif
		return err;
	}
	#else
	{
		#ifdef __SMP__
			lock_kernel();
		#endif
		k_filename = getname((char *) regs.ebx);
		err = PTR_ERR(k_filename);
		if (IS_ERR(filename))
			goto out;
		err = do_execve_(k_filename, (char **) regs.ecx, (char **) regs.edx, &regs);
		if (err == 0)
			current->flags &= ~PF_DTRACE;
		putname(k_filename);
out:
		#ifdef __SMP__
			unlock_kernel();
		#endif
		return err;
	}
	#endif
}


asmlinkage ssize_t new_sys_write(unsigned int fd, const char * buf, size_t count)
{
	struct file * file = NULL;
	int	err;

	if (!monitor_started) goto Done;
	
	if ((file=fget(fd)) && KAVMON_IS_WRITTEN (file->f_flags) && S_ISREG(file->f_dentry->d_inode->i_mode)) {
		file_t			* write_file;
		struct list_head        * f_list;
		int			  found;

		read_lock (&written_files_lock);

		found = 0;
		list_for_each(f_list, &written_files) {
			write_file = list_entry(f_list,file_t,list);
		
			if (write_file->file == file && write_file->tsk == current) {
				found = 1; break;

			}
		}
		
		read_unlock (&written_files_lock);


		if (!found) {
			
			write_file = MEM_ALLOC(sizeof(file_t));
			write_file->file	= file;
			write_file->tsk		= current;

			write_lock (&written_files_lock);
			list_add(&write_file->list,&written_files);
			write_unlock (&written_files_lock);

			goto Done;
		}
	}


	if (file) fput(file);

Done:
	err = old_sys_write(fd,buf, count);
	
	return err;
}


asmlinkage int new_sys_close (unsigned int fd)
{
	int err, flags = 0;
	char		* k_filename = NULL;
	struct file	* file;
	file_info	  f_info;


	if (!monitor_started) {
		
		err = old_sys_close (fd);

		return err;
	}

	if (Process_trusted(current)) {

		file = fget(fd);
		
		if (file) {
			file_really_written(file);
			fput(file);
		}

		err = old_sys_close (fd);
		
		return err;
	}


	if ( (file = fget(fd)) != NULL) {

		if ( file_really_written(file)) {
			k_filename = DENTRY_PATH(file);
			f_info.dev = kdev_t_to_nr(file->f_dentry->d_inode->i_rdev);
			f_info.ino = file->f_dentry->d_inode->i_ino;
			flags = file->f_flags;
		}
		
		fput (file);		
	}

	err = old_sys_close (fd);

	if (err == 0) {

		if ( k_filename && monitor_started) {
			int	err;
			
		err = Monitor_kernel_check_file(current->uid,current->pid,k_filename,
					&f_info,
					MONITOR_CLOSE_FILE,
					flags, &err);
					
			}
	}

	if ( k_filename )
		MEM_FREE (k_filename);

	return err;
}


asmlinkage int new_sys_unlink(const char * u_file)
{
	int err;
	char * k_filename = NULL;
	file_info	  f_info;


	if (!monitor_started) {

		err = old_sys_unlink(u_file);

		return err;
	}

	get_file_from_userspace (u_file, &f_info, &k_filename);

	if (k_filename && Monitor_kernel_trusted_1(&f_info)) {
	    MEM_FREE(k_filename);
	    return -EPERM;
	}

	err = 0;
	
	if (k_filename) {
		if (!err && !Process_trusted(current)) {
			err = Monitor_kernel_check_file(current->uid,current->pid,
							k_filename,
							&f_info,
							MONITOR_UNLINK_FILE,0, &err);
		}

		MEM_FREE (k_filename);
	}
	
	if (err) return err;

	err = old_sys_unlink(u_file);


	return err;
}

asmlinkage int new_sys_rename(const char *u_old_file, const char * u_new_file)
{
	int		  err;
	char		* k_old_filename;
	char		* k_new_filename;
	int		  trusted;
	file_info	  f_info_old, f_info_new;
	
	trusted = 0;
	err	= 0;

	if (!monitor_started) {

		err = old_sys_rename(u_old_file,u_new_file);
		
		return err;
	}

	get_file_from_userspace (u_old_file, &f_info_old, &k_old_filename);
	get_file_from_userspace (u_new_file, &f_info_new, &k_new_filename);


	if (k_old_filename && Monitor_kernel_trusted_1(&f_info_old)) {
		MEM_FREE(k_old_filename);
		if (k_new_filename) MEM_FREE (k_new_filename);
		return -EPERM;
	}

	if (k_new_filename && Monitor_kernel_trusted_1(&f_info_new)) {
		MEM_FREE(k_new_filename);
		if (k_old_filename) MEM_FREE (k_old_filename);
		return -EPERM;
	}

	trusted = Process_trusted(current);

	if (k_old_filename && !trusted) {
		err = Monitor_kernel_check_file(current->uid,current->pid,
						k_old_filename,
						&f_info_old,
						MONITOR_UNLINK_FILE,0,&err);
	}
	if (k_new_filename && !trusted) {
		err = Monitor_kernel_check_file(current->uid,current->pid,
						k_new_filename,
						&f_info_new,
						MONITOR_RENAME_FILE,0, &err);
	}

	if (k_old_filename) MEM_FREE (k_old_filename);
	if (k_new_filename) MEM_FREE (k_new_filename);

	if (err) return err;

	err = old_sys_rename(u_old_file,u_new_file);

	return err;
}


asmlinkage int new_sys_fork(struct pt_regs regs)
{
	int retval;

	retval = old_sys_fork(regs);

	if (retval>0) {

		Monitor_kernel_unregister((pid_t) retval);

		if (Monitor_kernel_trusted(current->pid, get_pgrp(current))) {
			file_info f_info;

			get_exec_info(current,&f_info);
			Monitor_kernel_register((pid_t) retval,&f_info);
		}
		
	}
		
	return retval;	
}

asmlinkage long new_sys_exit(int error_code)
{
	
	if (!monitor_started) {

		return old_sys_exit(error_code);
	}


	__exit_tsk();
	

	return old_sys_exit(error_code);
}


#ifdef LINUX_KERNEL26
asmlinkage long new_sys_exit_group(int error_code)
{

	if (!monitor_started) {

		return old_sys_exit_group(error_code);
	}

	__exit_tsk();
	

	return old_sys_exit_group(error_code);
}
	
#endif

void __exit_tsk(void)
{
	struct list_head        * f_list;
	file_t			* check_file;
	int			  trusted, num_proc;
	file_info		  f_info;
	pid_t			  grp;
	struct	task_struct	* p;

	trusted = Process_trusted(current);

again:	
	write_lock (&written_files_lock);

	list_for_each(f_list, &written_files) {
		check_file = list_entry(f_list,file_t,list);
		if (check_file->tsk == current) {
			char * filename;
			int	err;
			
			list_del(&check_file->list);
			write_unlock (&written_files_lock);

			if (monitor_started && !trusted) {
				filename = DENTRY_PATH(check_file->file);
				f_info.dev = kdev_t_to_nr(check_file->file->f_dentry->d_inode->i_rdev);
				f_info.ino = check_file->file->f_dentry->d_inode->i_ino;

				if (filename) {
					Monitor_kernel_check_file(current->uid,
								current->pid,filename,
								&f_info,
								MONITOR_CLOSE_FILE, check_file->file->f_flags, &err);
					MEM_FREE(filename);
				}
			}

			fput(check_file->file);
			MEM_FREE(check_file);
			goto again;
		}
	}
		
	write_unlock (&written_files_lock);

	Monitor_kernel_unregister(current->pid);

	read_lock(&tasklist_lock);

	grp = get_pgrp(current);
	num_proc = 0;

	for_each_process(p) {

		if (p->pid != current->pid && grp == get_pgrp(p)) num_proc++;

	}
	
	read_unlock(&tasklist_lock);

	if (!num_proc) Monitor_kernel_unregister_group(grp);
	
}


asmlinkage long new_sys_setpgid(pid_t pid, pid_t pgid)
{
	int err;

	err = old_sys_setpgid(pid,pgid);

	if (!err) {
		struct task_struct * tsk;

		if (!pid) pid = current->pid;
		
		read_lock(&tasklist_lock);
		tsk = find_task_by_pid(pid); 

		if (tsk) {
			Monitor_kernel_unregister_group(get_pgrp(tsk));
			Monitor_kernel_unregister(tsk->pid);
		}

		read_unlock(&tasklist_lock);
	}

	return err;
}

