
/*  
 * $Id: interface_monitor.c,v 1.43 2005/03/18 12:31:35 isv Exp $
 *
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2004
 * See License.txt for details
 *
 */



#include "module.h"

/* monitor functions */
static int	monitor_open 	( struct inode * inode, struct file * filp );
static int	monitor_close	( struct inode * inode, struct file * filp );
static int      monitor_ioctl (struct inode * inode, struct file *filp, unsigned int cmd,
				 unsigned long addr);

/* monitor daemon functions */
static ssize_t	monitor_daemon_read(struct file * file, char * ubuf, size_t len, loff_t * offset );
static ssize_t	monitor_daemon_write(struct file * file, const char * ubuf, size_t len, loff_t * offset );
static int	monitor_daemon_close(struct inode * inode, struct file * filp );
static unsigned int monitor_daemon_poll(struct file *file, poll_table * wait );
static int monitor_daemon_ioctl (struct inode * inode, struct file *filp, unsigned int cmd, unsigned long addr);



struct file_operations kavmonitor_file_operations = 
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,27))
	owner:		THIS_MODULE,
#endif			
	open:		monitor_open,
	ioctl:		monitor_ioctl,
	release:	monitor_close
};

static struct file_operations kavdaemon_file_operations = 
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,27))
	owner:		THIS_MODULE,
#endif			
	open:		monitor_open,
	ioctl:		monitor_daemon_ioctl,
	release:	monitor_daemon_close,
	poll:		monitor_daemon_poll,
	read:		monitor_daemon_read,
	write:		monitor_daemon_write
};

static int	monitor_daemon_pid;
static int	syscalls_table_altered;
int		monitor_started;		/* set if avir monitor is started */

int Monitor_module_init(void)
{
	monitor_daemon_pid = 0;
	syscalls_table_altered = 0;
	monitor_started = 0;
	
	return 0;
};

int Monitor_module_stop(TASK_INFO_)
{
	if (!monitor_daemon_pid) return 0;
	
	if (current->pid!=monitor_daemon_pid) return 0;

	Monitor_kernel_finit();
	if (!exploit_started) {
		if (Monitor_restore_calls()) syscalls_table_altered = 1;
	}
	
	Monitor_queue_cleanup(QUEUE_MONITOR);
	monitor_started = 0;

        return 0;
}


int monitor_open (struct inode * inode, struct file * filp)
{
	if (!monitor_daemon_pid) {
		file_info f_info;

		if (syscalls_table_altered) return -EFAULT;
		
		if (!exploit_started) {
		    Monitor_kernel_init();
		}

		get_exec_info(current,&f_info);
		Monitor_kernel_register(current->pid,&f_info);
		
		filp->f_op = &kavdaemon_file_operations;
		monitor_daemon_pid = current->pid;
		monitor_started = 1;
		if (!exploit_started) {
			Monitor_intercept_calls();
		}

		Monitor_ins_use();

	}


	return 0;
}

int monitor_close (struct inode * inode, struct file * filp)
{

	return 0;
}

int monitor_ioctl (struct inode * inode, struct file *filp, unsigned int cmd, unsigned long addr)
{
	char	* path;
	int	  trusted;
	int	  err_1;
	file_info f_info;

	switch (cmd) {
	
		case MONITOR_GROUP_CMD:
			if (Process_trusted(current)){
				(void) Monitor_kernel_register_group(get_pgrp(current));
			}
			return 0;

			break;
		
		case MONITOR_REGISTER_CMD:
		        path = get_exec_path(current,&f_info);
		
			if (Process_trusted(current)){
				trusted = 1;
				err_1	= 0;
			} else trusted = Monitor_kernel_check_file_sign(current->uid,current->pid,path, &err_1);
			
			if (!err_1 && trusted) {
				(void) Monitor_kernel_register(current->pid,&f_info);
				MEM_FREE(path);
				return  0;
			} else {
				MEM_FREE(path);
				if (err_1) return err_1;
				return -EINVAL;
			}
		break;
	}

	return -EINVAL; 	
}


ssize_t monitor_daemon_read (struct file * file, char * ubuf, size_t count, loff_t * offset)
{

	return Monitor_queue_read_interface (current, file, ubuf, count);
}


ssize_t monitor_daemon_write (struct file * file, const char * ubuf, size_t count, loff_t * offset)
{

	return Monitor_queue_write_interface (current, file, ubuf, count, offset);
}


int monitor_daemon_close (struct inode * inode, struct file * filp)
{
	if (current->pid!=monitor_daemon_pid) {

		return 0;
	}

	if (monitor_started) Monitor_module_stop(current);

	if (!syscalls_table_altered)  Monitor_dec_use();
	
	monitor_daemon_pid = 0;
	
	return 0;
}

unsigned int monitor_daemon_poll ( struct file *file, poll_table * wait )
{
	return Monitor_queue_poll_interface (current, file, wait);
}


int monitor_daemon_ioctl (struct inode * inode, struct file *filp, unsigned int cmd, unsigned long addr)
{
	switch (cmd) {
		
		case MONITOR_CONFIG_CMD:
			{
				monitor_config_ioctl * params;
				int			cache_size;

				params = (monitor_config_ioctl *) addr;
				if (!params) return -EFAULT;

				if (get_user(cache_size,&params->cache_size))
                                        return -EFAULT;
			
				m_config.cache_enable = cache_size?1:0;
				Monitor_cache_init(cache_size);

			}

				return 0;
	}
			
	return -EINVAL; 	
}


