/*  
 * $Id: queue.c,v 1.55 2005/04/15 08:57:50 isv Exp $
 *
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2004
 * See License.txt for details
 *
 */


#include "module.h"


#if  defined(__linux__)
static ssize_t queue_read_interface (TASK_INFO struct file * file, char * ubuf, size_t count,enum queue_entry_types queue_type);
static ssize_t queue_write_interface (TASK_INFO struct file * file, const char * ubuf, size_t count, loff_t * offset,
				      enum queue_entry_types queue_type);
static unsigned int queue_poll_interface (TASK_INFO struct file *file, poll_table * wait,enum queue_entry_types queue_type);

#elif defined(__FreeBSD__)
static ssize_t queue_read_interface (TASK_INFO dev_t dev, struct uio * uio, int ioflags,enum queue_entry_types queue_type);
static ssize_t queue_write_interface (TASK_INFO dev_t dev, struct uio *uio, int ioflag,enum queue_entry_types queue_type);
static unsigned int queue_poll_interface(TASK_INFO int events,enum queue_entry_types queue_type);

#endif

static int Monitor_queue_ref(mqueue_entry_t * queue);
static int Monitor_queue_rel(mqueue_entry_t * queue);



static lock_declare(queue_lock);

#ifdef __FreeBSD__
static struct selinfo		queue_empty_select;
#endif

/* Sleep the process until there is a free slot in the queue */
static wait_queue_head_t 	queue_busy_waitq;
/* Sleep the process until there is at least one item in the queue */
static wait_queue_head_t 	queue_empty_waitq;
static KLIST_HEAD(monitor_queue);
static u_short	last_queue_id;


int Monitor_queue_init(void)
{
	
	KINIT_LIST_HEAD(&monitor_queue);

	init_waitqueue_head (&queue_busy_waitq);
   	init_waitqueue_head (&queue_empty_waitq);

	lock_init(&queue_lock,"kavmonitor queue");
	
#if defined (__FreeBSD_version) && __FreeBSD_version < 500000
	queue_empty_select.si_flags = 0;
	queue_empty_select.si_pid = 0;
#endif	

	last_queue_id = 0;

	return 0;
}

void Monitor_queue_cleanup(enum queue_entry_types queue_type)
{
	struct list_head        * q_list;
	mqueue_entry_t		* queue;


	/* Wake up all processes in the queue */
	write_lock (&queue_lock);

	list_for_each(q_list, &monitor_queue) {
		queue = list_entry(q_list,mqueue_entry_t,list);
		if (queue->flags && queue->type == queue_type) {
			wake_up_interruptible (&queue->waitq);
		}

	}
		
	write_unlock (&queue_lock);
	wake_up_interruptible (&queue_busy_waitq);
	
}


int Monitor_queue_get_id(void)
{
	u_short			  queue_id;
	struct list_head        * q_list;
	mqueue_entry_t		* queue;

	read_lock (&queue_lock);
again:	
	queue_id = last_queue_id++;

	list_for_each(q_list, &monitor_queue) {
		queue = list_entry(q_list,mqueue_entry_t,list);
		if (queue->queue_id == queue_id) goto again;
	}

	read_unlock (&queue_lock);

	return queue_id;	
}

void Monitor_queue_set_answer(u_short queue_id, long answer)
{
	struct list_head        * q_list;
	mqueue_entry_t		* queue;

	read_lock (&queue_lock);

	list_for_each(q_list, &monitor_queue) {
		queue = list_entry(q_list,mqueue_entry_t,list);
		if (queue->queue_id == queue_id) {
			
			Monitor_queue_ref(queue);
	
			queue->answer = answer;
			
			read_unlock (&queue_lock);
			Monitor_queue_rel(queue);
			
			return;

		}

	}
		
	read_unlock (&queue_lock);
}


int Monitor_queue_add(enum queue_entry_types queue_type, 
			uid_t uid,pid_t pid, char * request_data, 
			int request_len, long * answer, u_short queue_id)
{
	mqueue_entry_t		* queue;

   	/* Add a new request */
#ifdef __linux__
	wait_queue_t		  wait;
#endif

	queue = MEM_ALLOC(sizeof(mqueue_entry_t));

	write_lock (&queue_lock);

	if ((!monitor_started && queue_type == QUEUE_MONITOR) || (!exploit_started && queue_type==QUEUE_EXPLOIT)) {

		write_unlock (&queue_lock);
	
		MEM_FREE(request_data);
		MEM_FREE(queue);
		*answer=MONITOR_ACCESS_ACCEPT;
		
		return 0;
	}
	

	queue->queue_id		= queue_id;
	queue->type		= queue_type;
	queue->request		= MEM_ALLOC(sizeof(mqueue_data_t));
	queue->request->data	= request_data;
	queue->request->len	= request_len;
	queue->request->left	= queue->request->len;
	queue->answer		= *answer;
	queue->devc		= 0;
	lock_init(&queue->lock,"queue item lock");
	queue->refcnt		= 1;


	queue->pid	= pid;
	queue->uid	= uid;
	queue->flags	= O_MQUEUE_READY;
	init_waitqueue_head(&queue->waitq);
	list_add(&queue->list,&monitor_queue);
	
#ifdef __linux__	
	
	init_waitqueue_entry(&wait, current);
	add_wait_queue(&queue->waitq, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	write_unlock (&queue_lock);

	Monitor_ins_use();
	
	wake_up(&queue_empty_waitq);

	schedule();

	remove_wait_queue(&queue->waitq, &wait);
	*answer = queue->answer;

	Monitor_queue_rel(queue);

	Monitor_dec_use();


	if (signal_pending(current)) {
		return -ERESTARTSYS;
	}

#endif	

	/* Indicate new data availability to the interface process */
#ifdef __FreeBSD__
	{
		int err;
		
		selwakeup (&queue_empty_select); /* Wake up also all select() calls*/ 
		write_unlock (&queue_lock);

		Monitor_ins_use();

		err = interruptible_sleep_on(&queue->waitq);
	
		Monitor_dec_use();

		*answer = queue->answer;
		Monitor_queue_rel(queue);

		if (err) return err;
	}
#endif	

	return 0;
}

void Monitor_queue_wakeup(u_short queue_id)
{
	struct list_head        * q_list;
	mqueue_entry_t		* queue;

	read_lock (&queue_lock);

	list_for_each(q_list, &monitor_queue) {
		queue = list_entry(q_list,mqueue_entry_t,list);
		if (queue->queue_id == queue_id) {
			read_unlock (&queue_lock);

			wake_up_interruptible (&queue->waitq);
			return;
		}

	}
	read_unlock (&queue_lock);

}

int Monitor_queue_ref(mqueue_entry_t * queue)
{
	int i;

	write_lock(&queue->lock);

	if (!queue->refcnt) {
		PRN("%s: called with cnt=0",__FUNCTION__);
	}

	i = ++queue->refcnt;

	write_unlock(&queue->lock);

	return i;
}

int Monitor_queue_rel(mqueue_entry_t * queue)
{
	int i;

	write_lock(&queue_lock);
	write_lock(&queue->lock);


	if (!queue->refcnt) {
		PRN("%s: called with cnt=0",__FUNCTION__);
	}
	
	i = --queue->refcnt;

	if (i!=0) {
		write_unlock(&queue_lock);
		write_unlock(&queue->lock);

		return i;
	}

	list_del(&queue->list);

	write_unlock(&queue_lock);
	write_unlock(&queue->lock);

	MEM_FREE(queue->request->data);
	MEM_FREE(queue->request);
	MEM_FREE(queue);

	return i;
}




#if  defined(__linux__)
/* monitor i/o functions */
ssize_t Monitor_queue_read_interface (TASK_INFO struct file * file, char * ubuf, size_t count)
{
	return queue_read_interface(TASK file,ubuf,count,QUEUE_MONITOR);


}
ssize_t Monitor_queue_write_interface (TASK_INFO struct file * file, const char * ubuf, size_t count, loff_t * offset)
{

	return queue_write_interface (TASK file,ubuf,count,offset,QUEUE_MONITOR);
}

unsigned int Monitor_queue_poll_interface (TASK_INFO struct file *file, poll_table * wait)
{

	return queue_poll_interface(TASK file,wait,QUEUE_MONITOR);

}

/* exploits checker i/o functions */

ssize_t Exploit_queue_read_interface (TASK_INFO struct file * file, char * ubuf, size_t count)
{
	return queue_read_interface(TASK file,ubuf,count,QUEUE_EXPLOIT);

}

ssize_t Exploit_queue_write_interface (TASK_INFO struct file * file, const char * ubuf, size_t count, loff_t * offset)
{

	return queue_write_interface (TASK file,ubuf,count,offset,QUEUE_EXPLOIT);
}

unsigned int Exploit_queue_poll_interface (TASK_INFO struct file *file, poll_table * wait)
{

	return queue_poll_interface(TASK file,wait,QUEUE_EXPLOIT);

}

#elif defined(__FreeBSD__)

/* monitor i/o functions */

ssize_t Monitor_queue_read_interface (TASK_INFO dev_t dev, struct uio * uio, int ioflags )
{

	return queue_read_interface(TASK dev, uio,ioflags,QUEUE_MONITOR);
}

ssize_t Monitor_queue_write_interface (TASK_INFO dev_t dev, struct uio *uio, int ioflag)
{

	return queue_write_interface(TASK dev,uio, ioflag,QUEUE_MONITOR);
}
unsigned int Monitor_queue_poll_interface(TASK_INFO int events)
{

	return queue_poll_interface(TASK events,QUEUE_MONITOR);
}


/* exploits checker i/o functions */
ssize_t Exploit_queue_read_interface (TASK_INFO dev_t dev, struct uio * uio, int ioflags )
{

	return queue_read_interface(TASK dev, uio,ioflags,QUEUE_EXPLOIT);
}

ssize_t Exploit_queue_write_interface (TASK_INFO dev_t dev, struct uio *uio, int ioflag)
{

	return queue_write_interface(TASK dev, uio,ioflag,QUEUE_EXPLOIT);
}
unsigned int Exploit_queue_poll_interface(TASK_INFO int events)
{

	return queue_poll_interface(TASK events,QUEUE_EXPLOIT);
}

#endif


#if  defined(__linux__)
ssize_t queue_read_interface (TASK_INFO struct file * file, char * ubuf, size_t count,enum queue_entry_types queue_type)
#elif defined(__FreeBSD__)
ssize_t queue_read_interface (TASK_INFO dev_t dev, struct uio * uio, int ioflags,enum queue_entry_types queue_type)
#endif
{
	int len = 0;
	mqueue_data_t		* request;
	mqueue_entry_t		* queue;
	struct list_head        * q_list;
#if defined(__linux__)
	dev_t dev = file->f_dentry->d_inode->i_rdev;
#endif	

	while ( 1 ) {
		mqueue_entry_t	* tmp_queue;
		
		write_lock (&queue_lock);

		queue		= NULL;
		
		list_for_each(q_list, &monitor_queue) {
			tmp_queue = list_entry(q_list,mqueue_entry_t,list);
			if (tmp_queue->flags & O_MQUEUE_READY && tmp_queue->type == queue_type) {

				if (tmp_queue->devc == dev) {
					queue = tmp_queue;
					break;
				}
				
				if (!queue && tmp_queue->devc == 0) {
					queue = tmp_queue;
				}
			}

			
		}

		if (queue) {
			if (!queue->devc) queue->devc = dev;
			break;
		}
		
		write_unlock (&queue_lock);			
			/* There is no data to sent - we sleep buddy if he wants to */
#if defined(__linux__)
		if ( (file->f_flags & O_NONBLOCK) != 0 )
			return -EAGAIN;

#elif defined(__FreeBSD__)
		if ( (ioflags & O_NONBLOCK) != 0 )
			return EAGAIN;

#endif

#ifdef __linux__
		Monitor_ins_use();
#endif
		interruptible_sleep_on (&queue_empty_waitq);

#ifdef __linux__			
		Monitor_dec_use();

		if ( signal_pending(current) )
			return -ERESTARTSYS;
#endif			
	}
	
	request = queue->request;

	Monitor_queue_ref(queue);
	
	write_unlock (&queue_lock);

#if defined(__linux__)
	len = MIN(count, request->left);
	
#elif defined(__FreeBSD__)
	len = MIN(uio->uio_resid, request->left);

#endif

#if defined(__linux__)
	if (copy_to_user (ubuf, (char *)(request->data) + (request->len - request->left) , len)) {

		Monitor_queue_rel(queue);
		return -EFAULT;

#elif defined(__FreeBSD__)
	if ( uiomove((char *)(request->data) + (request->len - request->left) , len,uio)){

		Monitor_queue_rel(queue);
		return EFAULT;
#endif	
	}

	request->left-=len;
	
	if ( request->left == 0 ){
		queue->flags &= ~O_MQUEUE_READY;
		queue->flags |= O_MQUEUE_TESTING;
	}
	

	Monitor_queue_rel(queue);

#if defined(__linux__)
	return len;
#elif defined(__FreeBSD__)	
	return 0;
#endif	
}

#if defined(__linux__)
ssize_t queue_write_interface (TASK_INFO struct file * file, const char * ubuf, 
			       size_t count, loff_t * offset,enum queue_entry_types queue_type)
#elif defined(__FreeBSD__)
ssize_t queue_write_interface (TASK_INFO dev_t dev, struct uio *uio, int ioflag,enum queue_entry_types queue_type)
#endif
{
	char * data;

#if defined (__FreeBSD__)
	int count;
	count = uio->uio_resid;
#endif	

	data = MEM_ALLOC(count +1);

#if defined(__linux__)
	if (copy_from_user(data,ubuf,count)) {
		MEM_FREE(data);
		return -EINVAL;

#elif defined(__FreeBSD__)
	if (uiomove (data, count, uio)!= 0) {
		MEM_FREE(data);
		return EINVAL;
#endif		
	}
	Monitor_kernel_parse(TASK data);
	MEM_FREE(data);

	return count;
}

#if defined(__linux__)
unsigned int queue_poll_interface (TASK_INFO  struct file *file, poll_table * wait,enum queue_entry_types queue_type)
{
	unsigned int		  ret = 0;
	mqueue_entry_t		* queue;
	struct list_head        * q_list;
    
	poll_wait(file, &queue_empty_waitq, wait);
	
	read_lock (&queue_lock);

	list_for_each(q_list, &monitor_queue) {
		queue = list_entry(q_list,mqueue_entry_t,list);
		if (queue->flags & O_MQUEUE_READY && queue->type == queue_type) {
			ret =  POLLIN | POLLRDNORM;
			break;
		}
	}

	read_unlock (&queue_lock);
	return ret;
}
#elif defined (__FreeBSD__)

unsigned int queue_poll_interface(TASK_INFO int events,enum queue_entry_types queue_type)
{	
	int revents = 0;
	mqueue_entry_t		* queue;
	struct list_head        * q_list;

	read_lock (&queue_lock);
	
	if ( events & (POLLIN | POLLRDNORM)) {
		list_for_each(q_list, &monitor_queue) {
			queue = list_entry(q_list,mqueue_entry_t,list);
			if (queue->flags & O_MQUEUE_READY && queue->type == queue_type) {
				revents |= events & (POLLIN | POLLRDNORM);
				break;
			}
		}

		if (!revents) 	selrecord (TASK &queue_empty_select);
	}

	if (events & (POLLOUT | POLLWRNORM))
		 revents |= (events & (POLLOUT | POLLWRNORM));

	read_unlock (&queue_lock);

	return revents;
}
#endif


